/* amd64.h - This is the interface definition for the AMD64
   architecture. */
/* Copyright (C) 2017 Robin Miyagi
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#ifndef ____AMD64_H____
#define ____AMD64_H____
/********************************************************************/
/*                                                                  */
/*			   Project Includes                         */
/*                                                                  */
/********************************************************************/
/********************************************************************/
/*                                                                  */
/*			   System Includes                          */
/*                                                                  */
/********************************************************************/
#include <stdint.h>
/********************************************************************/
/*                                                                  */
/*			  Macro Definitions                         */
/*                                                                  */
/********************************************************************/
/********************************************************************/
/*                                                                  */
/*			      Data Types                            */
/*                                                                  */
/********************************************************************/
/* Control Registers............................................... */
/*
 * Type definition and union for the CR0 register.  
 */
typedef union cr0_union cr0_T;
union cr0_union
{
  uint32_t i;
  struct
  {
    uint32_t pe:1;
    uint32_t mp:1;
    uint32_t em:1;
    uint32_t ts:1;
    uint32_t et:1;
    uint32_t ne:1;
    uint32_t reserved0:10;
    uint32_t wp:1;
    uint32_t reserved1:1;
    uint32_t am:1;
    uint32_t reserved2:10;
    uint32_t nw:1;
    uint32_t cd:1;
    uint32_t pg:1;
  } bits;
};
/* System Data Structures.......................................... */
/*
 * Type definition and union for a segment selector (SS).
 */
typedef union ss_union ss_T;
union ss_union
{
  uint16_t i;
  struct
  {
    uint16_t cpl:2;
    uint16_t ti:1;
    uint16_t si:13;
  } bits;
};
/*
 * Type definition and structure for a pseudodescriptor.
 */
typedef struct pseudo_descriptor_struct pseudo_descriptor_T;
struct pseudo_descriptor_struct
{
  uint16_t limit;
  void *base;
} __attribute__((__packed__));

/*  
 * Type definition and union for a descriptor (SD).
 */
typedef union desc_union desc_T;
union desc_union
{
  uint64_t i;
  /* Types of descriptors that belong in the GDT................... */
  /*
   * Code segment descriptor (CSD).  
   */
  struct
  {
    uint64_t segment_limit0:16;
    uint64_t base_address0:16;
    uint64_t base_address1:8;
    uint64_t a:1;
    uint64_t r:1;
    uint64_t c:1;
    uint64_t descriptor_type:2;	/* must be equal to DESCRIPTOR_TYPE_CODE. */
    uint64_t dpl:2;
    uint64_t p:1;
    uint64_t segment_limit1:4;
    uint64_t avl:1;
    uint64_t l:1;		/* must be zero in legacy mode. */
    uint64_t d:1;
    uint64_t g:1;
    uint64_t base_address2:8;
  } csd;
  /*
   * Data segment descriptor (DSD).  If the e bit is set to 1, then this
   * segment may be usable as a stack segment.
   */
  struct
  {
    uint64_t segment_limit0:16;
    uint64_t base_address0:16;
    uint64_t base_address1:8;
    uint64_t a:1;
    uint64_t w:1;
    uint64_t e:1;
    uint64_t descriptor_type:2;	/* must be equal to DESCRIPTOR_TYPE_DATA. */
    uint64_t dpl:2;
    uint64_t p:1;
    uint64_t segment_limit1:4;
    uint64_t avl:1;
    uint64_t l:1;
    uint64_t d_b:1;
    uint64_t g:1;
    uint64_t base_address2:8;
  } dsd;
};
/*		     Descriptor Macro Defintions                    */
/* 
 * Descriptor types. 
 */
#define DESCRIPTOR_TYPE_DATA 0x2
#define DESCRIPTOR_TYPE_CODE 0x3
/* 
 * Interrupt vectors. 
 */
/* Miscellaneous data types........................................ */
/*
 * Far pointer.  
 */
typedef struct far_pointer_struct far_pointer_T;
struct far_pointer_struct
{
  ss_T segment; 
  void *offset;
} __attribute__((__packed__));
/********************************************************************/
/*                                                                  */
/*			   Global Variables                         */
/*                                                                  */
/********************************************************************/
/********************************************************************/
/*                                                                  */
/*			 Interface Functions                        */
/*                                                                  */
/********************************************************************/
/********************************************************************/
/*                                                                  */
/*			   Inline Functions                         */
/*                                                                  */
/********************************************************************/
/* Control Register Access......................................... */
/*
 * Read and return the contents of the CR0 register.
 */
static inline cr0_T read_cr0 (void);
/* 
 * Write the contents of `cr0_register' to the CR0 register. 
 */
static inline void write_cr0 (cr0_T cr0_register);
/* Machine Data Structure Register Access.......................... */
/*
 * Write to global descriptor table
 */
static inline void write_gdtr (pseudo_descriptor_T *pd);
/* Miscellanious Machine Functions................................. */
/*
 * Set interrupts.  
 */
static inline void set_int (void);
/*  
 * Clear interrupts.
 */
static inline void clr_int (void);
/*  
 * Long jump.  Used for switching between real and protected mode.
 */
static inline void long_jump_32 (uint16_t seg, void *offset, void *parameter);

/********************************************************************/
/*                                                                  */
/*		    Inline Function Implementation                  */
/*                                                                  */
/********************************************************************/
static inline cr0_T
read_cr0 (void)
{
  cr0_T ret;

  asm ("movl %%cr0, %0"
       :"=r" (ret.i)
       : 
       :
       );
  return ret;
}
static inline void
write_cr0 (cr0_T cr0_register)
{
  asm ("movl %0, %%cr0"
       :
       : "r" (cr0_register.i)
       :
       );
}
static inline void
write_gdtr (pseudo_descriptor_T *pd)
{
  asm (".code32");
  asm ("lgdt %0"
       :
       : "m" (pd)
       :
       );
  asm (".code16gcc");
}
static inline void
write_seg_data (uint16_t index)
{
  ss_T selector;

  selector.i = 0;
  selector.bits.si = index;
  asm ("mov %0, %%ds"
       :
       : "r" (selector.i)
       :
       );
  asm ("mov %0, %%es"
       :
       : "r" (selector.i)
       :
       );
  asm ("mov %0, %%fs"
       :
       : "r" (selector.i)
       :
       );
  asm ("mov %0, %%gs"
       :
       : "r" (selector.i)
       :
       );
  asm ("mov %0, %%ss"
       :
       : "r" (selector.i)
	 :
       );
}
static inline void
set_int (void)
{
  asm ("sti");
}
static inline void
clr_int (void)
{
  asm ("cli");
}
static inline void
long_jump_32 (uint16_t seg, void *offset, void *parameter)
{
  cr0_T cr0_register;
  far_pointer_T fp;
  ss_T segment;

  segment.i = 0;
  segment.bits.si = seg;
  fp.offset = offset;
  fp.segment = segment;
  asm ("pushl %0"
       :
       : "r" (parameter)
       : 
       );
  cr0_register = read_cr0 ();
  cr0_register.bits.pe = 1;
  write_cr0 (cr0_register);
  asm (".code32; ljmp  *%0"
       :
       : "m" (fp)
       :
       );
}
#endif	/* ! ____AMD64_H____ */
