/* main.c - This is the 16 bit loader. */
/* Copyright (C) 2017 Robin Miyagi
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
/********************************************************************/
/*                                                                  */
/*			   Project Includes                         */
/*                                                                  */
/********************************************************************/
#include <console.h>
#include <drive.h>
#include <amd64.h>
#include <legacy/mit.h>
#include <legacy/selector32.h>
/********************************************************************/
/*                                                                  */
/*			   System Includes                          */
/*                                                                  */
/********************************************************************/
#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>
/********************************************************************/
/*                                                                  */
/*			  Macro Definitions                         */
/*                                                                  */
/********************************************************************/
/********************************************************************/
/*                                                                  */
/*			      Data Types                            */
/*                                                                  */
/********************************************************************/
/*
 * Type definition for loader "function"  
 */
typedef void (*loader_T) (void);
/*
 * Type definition and structure for load image table (LIT).
 */
typedef struct load_table_struct load_table_T;
struct load_table_struct
{
  loader_T entry;
  void *edata;
} __attribute__((packed)) ;
/********************************************************************/
/*                                                                  */
/*			   Global Variables                         */
/*                                                                  */
/********************************************************************/
/* 
 * Implemented in the linker script `linker.ld', this is the loader
 * information table.
 */
extern load_table_T lit;
/*
 * Implemented in the linker script `linker.ld', this is the data at
 * the beginning of the `.text' section.
 */
extern const char _start;
/*
 * Implemented in the linker script `linker.ld', this is the data at
 * the end of the `.data' section.
 */
extern const char _edata;
/********************************************************************/
/*                                                                  */
/*			   Module Variables                         */
/*                                                                  */
/********************************************************************/
/*
 * The machine information table (MIT).
 */
#define MAP_ENTRIES 0x10
static mit_T mt;
static mm_T map[MAP_ENTRIES];
/********************************************************************/
/*                                                                  */
/*			 Interface Functions                        */
/*                                                                  */
/********************************************************************/
/*
 * The main driver routine for boot sector.
 */
void cmain (void);
/********************************************************************/
/*                                                                  */
/*			   Module Functions                         */
/*                                                                  */
/********************************************************************/
/*
 * Setup the machine information table.  
 */
static void setup_mit (void);
/*
 * Print memory map.
 */
static void print_memory_map (void);
/*
 * Load the 32 bit loader code into memory.
 */
static void load32 (void);
/*
 * Set up the machine for the basic flat segmentation model.
 */
static void machine32 (void);
/*
 * Setup the global descriptor table (GDT).
 */
static void setup_gdt (void);
/*
 * Setup the interrupt descriptor table (IDT).
 */
static void setup_idt (void);
/********************************************************************/
/*                                                                  */
/*			   Inline Functions                         */
/*                                                                  */
/********************************************************************/
/********************************************************************/
/*                                                                  */
/*			    Implementation                          */
/*                                                                  */
/********************************************************************/
void
cmain (void)
{
  printf ("The value of cr0 is 0x%x.\r\n", read_cr0 ());
  setup_mit ();
  load32 ();
  printf ("Please press the `ENTER' key to continue...");
  getchar ();
  printf ("\r\n");

  machine32 ();
  long_jump_32 (CODE_DESC_INDEX, lit.entry, &mt);
  /*
   * Make sure that we do not run off of the end of the program.
   */
  error ("The loader should NOT ever return");
}
static void
setup_mit (void)
{
  uint16_t rtn = 0;
  uint32_t count = 0;
  uint32_t i;

  for (;;)
    {
      rtn = get_mm_entry (map, rtn);
      if (mm_e820_error)
	error ("int $0x15, %eax=0xe820 BIOS function is not supported");
      ++count;
      if (rtn == 0)
	break;
      if (count == 0)
	{
	  if (mm_version_error)
	    error ("Memory map version mismatch");
	}
    }
  if (count > MAP_ENTRIES)
    error ("The count 0x%x of map entries excedes the MAP_ENTRIES value of 0x%x.  "
	   "Please see the MAP_ENTRIES macro definition",
	   count, MAP_ENTRIES);
  mt.mbz = NULL;
  mt.memory_map = map;
  mt.map_count = count;
  for (i = 0; i < count; ++i)
    rtn = get_mm_entry (&map[i], rtn);
  print_memory_map ();
}
static void
print_memory_map (void)
{
  uint32_t i;
 
  for (i = 0; i < mt.map_count; ++i)
    {
      printf ("0x%x%x    0x%x%x    ",
	      mt.memory_map[i].base_address_high,
	      mt.memory_map[i].base_address_low,
	      mt.memory_map[i].length_high,
	      mt.memory_map[i].length_low
	      );
      switch (mt.memory_map[i].type)
	{
	case MM_REGION_TYPE_USABLE:
	  printf ("usable");
	  break;
	case MM_REGION_TYPE_RESERVED:
	  printf ("reserved");
	  break;
	case MM_REGION_TYPE_RECLAIMABLE:
	  printf ("reclaimable");
	  break;
	case MM_REGION_TYPE_NVS:
	  printf ("NVS");
	  break;
	case MM_REGION_TYPE_BAD:
	  printf ("**** bad ****");
	  break;
	default:
	  error ("\r\ninvalid memory region type found in memory map at index 0x%x", i);
	}
      printf ("\r\n");
    }
}
static void
load32 (void)
{
  uint32_t size16;	/* Size of 16 bit loader. */
  uint32_t size32;	/* Size of 32 bit loader. */
  
  size16 = sectors (&_start, &_edata);
  ++size16;	/* The 32 bit loader starts in the next block. */
  if (read_drive (size16, 1, &lit) == false)
    error ("unable to load the 32 bit loader information table");
  ++size16;	/* The rest of the 32 bit loader start in the next block. */
  size32 = sectors (&lit, lit.edata) - 1;
  printf ("0x%x sectors left to read.\r\n", size32);
  if (size32 != 0)
    {
      if (read_drive (size16, size32, (void *) &lit + SECTOR_SIZE) == false)
	error ("unable to load the rest of the 32 bit loader.  "
	       "It is recommended that you check the linker script "
	       "for the 32 bit linker code.");
    }
  printf ("The loadable portion of the 32 bit boot loader code occupies "
	  "memory addresses from 0x%x to 0x%x and the entry point is 0x%x.\r\n",
	  &lit, lit.edata, lit.entry);
}
static void
machine32 (void)
{
  clr_int ();
  setup_gdt ();
  setup_idt ();
}
static void
setup_gdt (void)
{
  __attribute__((__aligned__(8) )) static desc_T gdt[DESCRIPTOR_COUNT];
  static pseudo_descriptor_T pd;

  /*
   * set up the null descriptor.
   */
  gdt[NULL_DESC_INDEX].i = 0;
  /*  
   *  Setup the code descriptor --- flat segmentation.
   */
  gdt[CODE_DESC_INDEX].csd.descriptor_type = DESCRIPTOR_TYPE_CODE;
  gdt[CODE_DESC_INDEX].csd.base_address0 = 0;
  gdt[CODE_DESC_INDEX].csd.base_address1 = 0;
  gdt[CODE_DESC_INDEX].csd.base_address2 = 0;
  gdt[CODE_DESC_INDEX].csd.segment_limit0 = -1;
  gdt[CODE_DESC_INDEX].csd.segment_limit1 = -1;
  gdt[CODE_DESC_INDEX].csd.dpl = 0;
  gdt[CODE_DESC_INDEX].csd.p = 1;
  gdt[CODE_DESC_INDEX].csd.d = 1;
  gdt[CODE_DESC_INDEX].csd.g = 1;
  gdt[CODE_DESC_INDEX].csd.c = 1;
  gdt[CODE_DESC_INDEX].csd.r = 1;
  gdt[CODE_DESC_INDEX].csd.a = 0;
  gdt[CODE_DESC_INDEX].csd.l = 0;
  /*  
   * Setup the stack descriptor --- flat segmentation, expanding downards.
   */
  gdt[STACK_DESC_INDEX].dsd.descriptor_type = DESCRIPTOR_TYPE_DATA;
  gdt[STACK_DESC_INDEX].dsd.base_address0 = -1;
  gdt[STACK_DESC_INDEX].dsd.base_address1 = -1;
  gdt[STACK_DESC_INDEX].dsd.base_address2 = -1;
  gdt[STACK_DESC_INDEX].dsd.segment_limit0 = -1;
  gdt[STACK_DESC_INDEX].dsd.segment_limit1 = -1;
  gdt[STACK_DESC_INDEX].dsd.dpl = 0;
  gdt[STACK_DESC_INDEX].dsd.p = 1;
  gdt[STACK_DESC_INDEX].dsd.d_b = 1;
  gdt[STACK_DESC_INDEX].dsd.g = 1;
  gdt[STACK_DESC_INDEX].dsd.e = 1;
  gdt[STACK_DESC_INDEX].dsd.w = 1;
  gdt[STACK_DESC_INDEX].dsd.a = 0;
  gdt[STACK_DESC_INDEX].dsd.l = 0;
  /*  
   * Setup the data descriptor --- flat segmentation.
   */
  gdt[DATA_DESC_INDEX].dsd.descriptor_type = DESCRIPTOR_TYPE_DATA;
  gdt[DATA_DESC_INDEX].dsd.base_address0 = 0;
  gdt[DATA_DESC_INDEX].dsd.base_address1 = 0;
  gdt[DATA_DESC_INDEX].dsd.base_address2 = 0;
  gdt[DATA_DESC_INDEX].dsd.segment_limit0 = -1;
  gdt[DATA_DESC_INDEX].dsd.segment_limit1 = -1;
  gdt[DATA_DESC_INDEX].dsd.dpl = 0;
  gdt[DATA_DESC_INDEX].dsd.p = 1;
  gdt[DATA_DESC_INDEX].dsd.d_b = 1;
  gdt[DATA_DESC_INDEX].dsd.g = 1;
  gdt[DATA_DESC_INDEX].dsd.e = 0;
  gdt[DATA_DESC_INDEX].dsd.w = 1;
  gdt[DATA_DESC_INDEX].dsd.a = 0;
  gdt[DATA_DESC_INDEX].dsd.l = 0;
  /*
   * Prepare the pseudo descriptor and write the location of same to
   * the global descriptor table register (GDTR).
   */
  pd.limit = DESCRIPTOR_COUNT * sizeof (desc_T);
  pd.base = gdt;
  write_gdtr (&pd);
}
static void
setup_idt (void)
{
  
}
